@@ -17,10 +17,12 @@
 
 #include "CClientTaskBarReceiver.h"
 
+class CBufferedLogOutputter;
+
 //! Implementation of CClientTaskBarReceiver for X Windows
 class CXWindowsClientTaskBarReceiver : public CClientTaskBarReceiver {
 public:
-	CXWindowsClientTaskBarReceiver();
+	CXWindowsClientTaskBarReceiver(const CBufferedLogOutputter*);
 	virtual ~CXWindowsClientTaskBarReceiver();
 
 	// IArchTaskBarReceiver overrides
@@ -28,10 +30,6 @@ public:
 	virtual void		runMenu(int x, int y);
 	virtual void		primaryAction();
 	virtual const Icon	getIcon() const;
-
-protected:
-	// CClientTaskBarReceiver overrides
-	virtual void		onStatusChanged();
 };
 
 #endif
