@@ -28,13 +28,18 @@
 #include "CLog.h"
 #include "TMethodEventJob.h"
 #include "CArch.h"
+#include <string.h>
 
 //
 // CServer
 //
 
-CEvent::Type			CServer::s_errorEvent        = CEvent::kUnknown;
-CEvent::Type			CServer::s_disconnectedEvent = CEvent::kUnknown;
+CEvent::Type			CServer::s_errorEvent         = CEvent::kUnknown;
+CEvent::Type			CServer::s_connectedEvent     = CEvent::kUnknown;
+CEvent::Type			CServer::s_disconnectedEvent  = CEvent::kUnknown;
+CEvent::Type			CServer::s_switchToScreen     = CEvent::kUnknown;
+CEvent::Type			CServer::s_switchInDirection  = CEvent::kUnknown;
+CEvent::Type			CServer::s_lockCursorToScreen = CEvent::kUnknown;
 
 CServer::CServer(const CConfig& config, CPrimaryClient* primaryClient) :
 	m_primaryClient(primaryClient),
@@ -44,7 +49,8 @@ CServer::CServer(const CConfig& config, CPrimaryClient* primaryClient) :
 	m_yDelta(0),
 	m_xDelta2(0),
 	m_yDelta2(0),
-	m_config(config),
+	m_config(),
+	m_inputFilter(m_config.getInputFilter()),
 	m_activeSaver(NULL),
 	m_switchDir(kNoDirection),
 	m_switchScreen(NULL),
@@ -54,11 +60,12 @@ CServer::CServer(const CConfig& config, CPrimaryClient* primaryClient) :
 	m_switchTwoTapEngaged(false),
 	m_switchTwoTapArmed(false),
 	m_switchTwoTapZone(3),
-	m_relativeMoves(false)
+	m_relativeMoves(false),
+	m_lockedToScreen(false)
 {
 	// must have a primary client and it must have a canonical name
 	assert(m_primaryClient != NULL);
-	assert(m_config.isScreen(primaryClient->getName()));
+	assert(config.isScreen(primaryClient->getName()));
 
 	CString primaryName = getName(primaryClient);
 
@@ -79,23 +86,23 @@ CServer::CServer(const CConfig& config, CPrimaryClient* primaryClient) :
 							new TMethodEventJob<CServer>(this,
 								&CServer::handleSwitchWaitTimeout));
 	EVENTQUEUE->adoptHandler(IPlatformScreen::getKeyDownEvent(),
-							m_primaryClient->getEventTarget(),
+							m_inputFilter,
 							new TMethodEventJob<CServer>(this,
 								&CServer::handleKeyDownEvent));
 	EVENTQUEUE->adoptHandler(IPlatformScreen::getKeyUpEvent(),
-							m_primaryClient->getEventTarget(),
+							m_inputFilter,
 							new TMethodEventJob<CServer>(this,
 								&CServer::handleKeyUpEvent));
 	EVENTQUEUE->adoptHandler(IPlatformScreen::getKeyRepeatEvent(),
-							m_primaryClient->getEventTarget(),
+							m_inputFilter,
 							new TMethodEventJob<CServer>(this,
 								&CServer::handleKeyRepeatEvent));
 	EVENTQUEUE->adoptHandler(IPlatformScreen::getButtonDownEvent(),
-							m_primaryClient->getEventTarget(),
+							m_inputFilter,
 							new TMethodEventJob<CServer>(this,
 								&CServer::handleButtonDownEvent));
 	EVENTQUEUE->adoptHandler(IPlatformScreen::getButtonUpEvent(),
-							m_primaryClient->getEventTarget(),
+							m_inputFilter,
 							new TMethodEventJob<CServer>(this,
 								&CServer::handleButtonUpEvent));
 	EVENTQUEUE->adoptHandler(IPlatformScreen::getMotionOnPrimaryEvent(),
@@ -118,32 +125,51 @@ CServer::CServer(const CConfig& config, CPrimaryClient* primaryClient) :
 							m_primaryClient->getEventTarget(),
 							new TMethodEventJob<CServer>(this,
 								&CServer::handleScreensaverDeactivatedEvent));
+	EVENTQUEUE->adoptHandler(getSwitchToScreenEvent(),
+							m_inputFilter,
+							new TMethodEventJob<CServer>(this,
+								&CServer::handleSwitchToScreenEvent));
+	EVENTQUEUE->adoptHandler(getSwitchInDirectionEvent(),
+							m_inputFilter,
+							new TMethodEventJob<CServer>(this,
+								&CServer::handleSwitchInDirectionEvent));
+	EVENTQUEUE->adoptHandler(getLockCursorToScreenEvent(),
+							m_inputFilter,
+							new TMethodEventJob<CServer>(this,
+								&CServer::handleLockCursorToScreenEvent));
+	EVENTQUEUE->adoptHandler(IPlatformScreen::getFakeInputBeginEvent(),
+							m_inputFilter,
+							new TMethodEventJob<CServer>(this,
+								&CServer::handleFakeInputBeginEvent));
+	EVENTQUEUE->adoptHandler(IPlatformScreen::getFakeInputEndEvent(),
+							m_inputFilter,
+							new TMethodEventJob<CServer>(this,
+								&CServer::handleFakeInputEndEvent));
 
 	// add connection
 	addClient(m_primaryClient);
 
-	// process options locally
-	processOptions();
-
-	// tell primary client about its options
-	sendOptions(m_primaryClient);
+	// set initial configuration
+	setConfig(config);
 
+	// enable primary client
 	m_primaryClient->enable();
+	m_inputFilter->setPrimaryClient(m_primaryClient);
 }
 
 CServer::~CServer()
 {
 	// remove event handlers and timers
 	EVENTQUEUE->removeHandler(IPlatformScreen::getKeyDownEvent(),
-							m_primaryClient->getEventTarget());
+							m_inputFilter);
 	EVENTQUEUE->removeHandler(IPlatformScreen::getKeyUpEvent(),
-							m_primaryClient->getEventTarget());
+							m_inputFilter);
 	EVENTQUEUE->removeHandler(IPlatformScreen::getKeyRepeatEvent(),
-							m_primaryClient->getEventTarget());
+							m_inputFilter);
 	EVENTQUEUE->removeHandler(IPlatformScreen::getButtonDownEvent(),
-							m_primaryClient->getEventTarget());
+							m_inputFilter);
 	EVENTQUEUE->removeHandler(IPlatformScreen::getButtonUpEvent(),
-							m_primaryClient->getEventTarget());
+							m_inputFilter);
 	EVENTQUEUE->removeHandler(IPlatformScreen::getMotionOnPrimaryEvent(),
 							m_primaryClient->getEventTarget());
 	EVENTQUEUE->removeHandler(IPlatformScreen::getMotionOnSecondaryEvent(),
@@ -154,6 +180,10 @@ CServer::~CServer()
 							m_primaryClient->getEventTarget());
 	EVENTQUEUE->removeHandler(IPlatformScreen::getScreensaverDeactivatedEvent(),
 							m_primaryClient->getEventTarget());
+	EVENTQUEUE->removeHandler(IPlatformScreen::getFakeInputBeginEvent(),
+							m_inputFilter);
+	EVENTQUEUE->removeHandler(IPlatformScreen::getFakeInputEndEvent(),
+							m_inputFilter);
 	EVENTQUEUE->removeHandler(CEvent::kTimer, this);
 	stopSwitch();
 
@@ -168,7 +198,11 @@ CServer::~CServer()
 		delete client;
 	}
 
-	// disconnect primary client
+	// remove input filter
+	m_inputFilter->setPrimaryClient(NULL);
+
+	// disable and disconnect primary client
+	m_primaryClient->disable();
 	removeClient(m_primaryClient);
 }
 
@@ -188,6 +222,21 @@ CServer::setConfig(const CConfig& config)
 	m_config = config;
 	processOptions();
 
+	// add ScrollLock as a hotkey to lock to the screen.  this was a
+	// built-in feature in earlier releases and is now supported via
+	// the user configurable hotkey mechanism.  if the user has already
+	// registered ScrollLock for something else then that will win but
+	// we will unfortunately generate a warning.  if the user has
+	// configured a CLockCursorToScreenAction then we don't add
+	// ScrollLock as a hotkey.
+	if (!m_config.hasLockToScreenAction()) {
+		IPlatformScreen::CKeyInfo* key =
+			IPlatformScreen::CKeyInfo::alloc(kKeyScrollLock, 0, 0, 0);
+		CInputFilter::CRule rule(new CInputFilter::CKeystrokeCondition(key));
+		rule.adoptAction(new CInputFilter::CLockCursorToScreenAction, true);
+		m_inputFilter->addFilterRule(rule);
+	}
+
 	// tell primary screen about reconfiguration
 	m_primaryClient->reconfigure(getActivePrimarySides());
 
@@ -234,6 +283,12 @@ CServer::adoptClient(IClient* client)
 	if (m_activeSaver != NULL) {
 		client->screensaver(true);
 	}
+
+	// send notification
+	CServer::CScreenConnectedInfo* info =
+		CServer::CScreenConnectedInfo::alloc(getName(client));
+	EVENTQUEUE->addEvent(CEvent(CServer::getConnectedEvent(),
+								m_primaryClient->getEventTarget(), info));
 }
 
 void
@@ -272,6 +327,13 @@ CServer::getErrorEvent()
 							"CServer::error");
 }
 
+CEvent::Type
+CServer::getConnectedEvent()
+{
+	return CEvent::registerTypeOnce(s_connectedEvent,
+							"CServer::connected");
+}
+
 CEvent::Type
 CServer::getDisconnectedEvent()
 {
@@ -279,16 +341,25 @@ CServer::getDisconnectedEvent()
 							"CServer::disconnected");
 }
 
-bool
-CServer::onCommandKey(KeyID id, KeyModifierMask /*mask*/, bool /*down*/)
+CEvent::Type
+CServer::getSwitchToScreenEvent()
 {
-	if (id == kKeyScrollLock) {
-		m_primaryClient->reconfigure(getActivePrimarySides());
-		if (!isLockedToScreenServer()) {
-			stopRelativeMoves();
-		}
-	}
-	return false;
+	return CEvent::registerTypeOnce(s_switchToScreen,
+							"CServer::switchToScreen");
+}
+
+CEvent::Type
+CServer::getSwitchInDirectionEvent()
+{
+	return CEvent::registerTypeOnce(s_switchInDirection,
+							"CServer::switchInDirection");
+}
+
+CEvent::Type
+CServer::getLockCursorToScreenEvent()
+{
+	return CEvent::registerTypeOnce(s_lockCursorToScreen,
+							"CServer::lockCursorToScreen");
 }
 
 CString
@@ -306,16 +377,16 @@ CServer::getActivePrimarySides() const
 {
 	UInt32 sides = 0;
 	if (!isLockedToScreenServer()) {
-		if (getNeighbor(m_primaryClient, kLeft) != NULL) {
+		if (hasAnyNeighbor(m_primaryClient, kLeft)) {
 			sides |= kLeftMask;
 		}
-		if (getNeighbor(m_primaryClient, kRight) != NULL) {
+		if (hasAnyNeighbor(m_primaryClient, kRight)) {
 			sides |= kRightMask;
 		}
-		if (getNeighbor(m_primaryClient, kTop) != NULL) {
+		if (hasAnyNeighbor(m_primaryClient, kTop)) {
 			sides |= kTopMask;
 		}
-		if (getNeighbor(m_primaryClient, kBottom) != NULL) {
+		if (hasAnyNeighbor(m_primaryClient, kBottom)) {
 			sides |= kBottomMask;
 		}
 	}
@@ -326,7 +397,7 @@ bool
 CServer::isLockedToScreenServer() const
 {
 	// locked if scroll-lock is toggled on
-	return ((m_primaryClient->getToggleMask() & KeyModifierScrollLock) != 0);
+	return m_lockedToScreen;
 }
 
 bool
@@ -334,7 +405,7 @@ CServer::isLockedToScreen() const
 {
 	// locked if we say we're locked
 	if (isLockedToScreenServer()) {
-		LOG((CLOG_DEBUG "locked by ScrollLock"));
+		LOG((CLOG_DEBUG "locked to screen"));
 		return true;
 	}
 
@@ -428,8 +499,76 @@ CServer::switchScreen(IClient* dst, SInt32 x, SInt32 y, bool forScreensaver)
 	}
 }
 
+void
+CServer::jumpToScreen(IClient* newScreen)
+{
+	assert(newScreen != NULL);
+
+	// warp to the center of the new client
+	// XXX -- would be better to save and restore the last position
+	SInt32 ax, ay, aw, ah;
+	newScreen->getShape(ax, ay, aw, ah);
+	SInt32 x = ax + (aw >> 1);
+	SInt32 y = ay + (ah >> 1);
+	
+	switchScreen(newScreen, x, y, false);
+}
+
+float
+CServer::mapToFraction(IClient* client,
+				EDirection dir, SInt32 x, SInt32 y) const
+{
+	SInt32 sx, sy, sw, sh;
+	client->getShape(sx, sy, sw, sh);
+	switch (dir) {
+	case kLeft:
+	case kRight:
+		return static_cast<float>(y - sy + 0.5f) / static_cast<float>(sh);
+
+	case kTop:
+	case kBottom:
+		return static_cast<float>(x - sx + 0.5f) / static_cast<float>(sw);
+
+	case kNoDirection:
+		assert(0 && "bad direction");
+		break;
+	}
+	return 0.0f;
+}
+
+void
+CServer::mapToPixel(IClient* client,
+				EDirection dir, float f, SInt32& x, SInt32& y) const
+{
+	SInt32 sx, sy, sw, sh;
+	client->getShape(sx, sy, sw, sh);
+	switch (dir) {
+	case kLeft:
+	case kRight:
+		y = static_cast<SInt32>(f * sh) + sy;
+		break;
+
+	case kTop:
+	case kBottom:
+		x = static_cast<SInt32>(f * sw) + sx;
+		break;
+
+	case kNoDirection:
+		assert(0 && "bad direction");
+		break;
+	}
+}
+
+bool
+CServer::hasAnyNeighbor(IClient* client, EDirection dir) const
+{
+	assert(client != NULL);
+
+	return m_config.hasNeighbor(getName(client), dir);
+}
+
 IClient*
-CServer::getNeighbor(IClient* src, EDirection dir) const
+CServer::getNeighbor(IClient* src, EDirection dir, SInt32& x, SInt32& y) const
 {
 	// note -- must be locked on entry
 
@@ -440,21 +579,19 @@ CServer::getNeighbor(IClient* src, EDirection dir) const
 	assert(!srcName.empty());
 	LOG((CLOG_DEBUG2 "find neighbor on %s of \"%s\"", CConfig::dirName(dir), srcName.c_str()));
 
-	// get first neighbor.  if it's the source then the source jumps
-	// to itself and we return the source.
-	CString dstName(m_config.getNeighbor(srcName, dir));
-	if (dstName == srcName) {
-		LOG((CLOG_DEBUG2 "\"%s\" is on %s of \"%s\"", dstName.c_str(), CConfig::dirName(dir), srcName.c_str()));
-		return src;
-	}
+	// convert position to fraction
+	float t = mapToFraction(src, dir, x, y);
 
-	// keep checking
+	// search for the closest neighbor that exists in direction dir
+	float tTmp;
 	for (;;) {
+		CString dstName(m_config.getNeighbor(srcName, dir, t, &tTmp));
+
 		// if nothing in that direction then return NULL. if the
 		// destination is the source then we can make no more
 		// progress in this direction.  since we haven't found a
 		// connected neighbor we return NULL.
-		if (dstName.empty() || dstName == srcName) {
+		if (dstName.empty()) {
 			LOG((CLOG_DEBUG2 "no neighbor on %s of \"%s\"", CConfig::dirName(dir), srcName.c_str()));
 			return NULL;
 		}
@@ -463,7 +600,8 @@ CServer::getNeighbor(IClient* src, EDirection dir) const
 		// ready then we can stop.
 		CClientList::const_iterator index = m_clients.find(dstName);
 		if (index != m_clients.end()) {
-			LOG((CLOG_DEBUG2 "\"%s\" is on %s of \"%s\"", dstName.c_str(), CConfig::dirName(dir), srcName.c_str()));
+			LOG((CLOG_DEBUG2 "\"%s\" is on %s of \"%s\" at %f", dstName.c_str(), CConfig::dirName(dir), srcName.c_str(), t));
+			mapToPixel(index->second, dir, tTmp, x, y);
 			return index->second;
 		}
 
@@ -471,13 +609,13 @@ CServer::getNeighbor(IClient* src, EDirection dir) const
 		LOG((CLOG_DEBUG2 "ignored \"%s\" on %s of \"%s\"", dstName.c_str(), CConfig::dirName(dir), srcName.c_str()));
 		srcName = dstName;
 
-		// look up name of neighbor of skipped screen
-		dstName = m_config.getNeighbor(srcName, dir);
+		// use position on skipped screen
+		t = tTmp;
 	}
 }
 
 IClient*
-CServer::getNeighbor(IClient* src,
+CServer::mapToNeighbor(IClient* src,
 				EDirection srcSide, SInt32& x, SInt32& y) const
 {
 	// note -- must be locked on entry
@@ -485,16 +623,14 @@ CServer::getNeighbor(IClient* src,
 	assert(src != NULL);
 
 	// get the first neighbor
-	IClient* dst = getNeighbor(src, srcSide);
+	IClient* dst = getNeighbor(src, srcSide, x, y);
 	if (dst == NULL) {
 		return NULL;
 	}
 
-	// get the source screen's size (needed for kRight and kBottom)
-	SInt32 sx, sy, sw, sh;
+	// get the source screen's size
 	SInt32 dx, dy, dw, dh;
 	IClient* lastGoodScreen = src;
-	lastGoodScreen->getShape(sx, sy, sw, sh);
 	lastGoodScreen->getShape(dx, dy, dw, dh);
 
 	// find destination screen, adjusting x or y (but not both).  the
@@ -513,7 +649,7 @@ CServer::getNeighbor(IClient* src,
 				break;
 			}
 			LOG((CLOG_DEBUG2 "skipping over screen %s", getName(dst).c_str()));
-			dst = getNeighbor(lastGoodScreen, srcSide);
+			dst = getNeighbor(lastGoodScreen, srcSide, x, y);
 		}
 		assert(lastGoodScreen != NULL);
 		x += dx;
@@ -529,7 +665,7 @@ CServer::getNeighbor(IClient* src,
 				break;
 			}
 			LOG((CLOG_DEBUG2 "skipping over screen %s", getName(dst).c_str()));
-			dst = getNeighbor(lastGoodScreen, srcSide);
+			dst = getNeighbor(lastGoodScreen, srcSide, x, y);
 		}
 		assert(lastGoodScreen != NULL);
 		x += dx;
@@ -545,7 +681,7 @@ CServer::getNeighbor(IClient* src,
 				break;
 			}
 			LOG((CLOG_DEBUG2 "skipping over screen %s", getName(dst).c_str()));
-			dst = getNeighbor(lastGoodScreen, srcSide);
+			dst = getNeighbor(lastGoodScreen, srcSide, x, y);
 		}
 		assert(lastGoodScreen != NULL);
 		y += dy;
@@ -557,11 +693,11 @@ CServer::getNeighbor(IClient* src,
 			y -= dh;
 			lastGoodScreen = dst;
 			lastGoodScreen->getShape(dx, dy, dw, dh);
-			if (y < sh) {
+			if (y < dh) {
 				break;
 			}
 			LOG((CLOG_DEBUG2 "skipping over screen %s", getName(dst).c_str()));
-			dst = getNeighbor(lastGoodScreen, srcSide);
+			dst = getNeighbor(lastGoodScreen, srcSide, x, y);
 		}
 		assert(lastGoodScreen != NULL);
 		y += dy;
@@ -580,88 +716,61 @@ CServer::getNeighbor(IClient* src,
 	// to avoid the jump zone.  if entering a side that doesn't have
 	// a neighbor (i.e. an asymmetrical side) then we don't need to
 	// move inwards because that side can't provoke a jump.
-	if (dst == m_primaryClient) {
-		const CString dstName(getName(dst));
-		switch (srcSide) {
-		case kLeft:
-			if (!m_config.getNeighbor(dstName, kRight).empty() &&
-				x > dx + dw - 1 - getJumpZoneSize(dst))
-				x = dx + dw - 1 - getJumpZoneSize(dst);
-			break;
-
-		case kRight:
-			if (!m_config.getNeighbor(dstName, kLeft).empty() &&
-				x < dx + getJumpZoneSize(dst))
-				x = dx + getJumpZoneSize(dst);
-			break;
-
-		case kTop:
-			if (!m_config.getNeighbor(dstName, kBottom).empty() &&
-				y > dy + dh - 1 - getJumpZoneSize(dst))
-				y = dy + dh - 1 - getJumpZoneSize(dst);
-			break;
-
-		case kBottom:
-			if (!m_config.getNeighbor(dstName, kTop).empty() &&
-				y < dy + getJumpZoneSize(dst))
-				y = dy + getJumpZoneSize(dst);
-			break;
-
-		case kNoDirection:
-			assert(0 && "bad direction");
-			return NULL;
-		}
-	}
-
-	// adjust the coordinate orthogonal to srcSide to account for
-	// resolution differences.  for example, if y is 200 pixels from
-	// the top on a screen 1000 pixels high (20% from the top) when
-	// we cross the left edge onto a screen 600 pixels high then y
-	// should be set 120 pixels from the top (again 20% from the
-	// top).
-	switch (srcSide) {
-	case kLeft:
-	case kRight:
-		y -= sy;
-		if (y < 0) {
-			y = 0;
-		}
-		else if (y >= sh) {
-			y = dh - 1;
-		}
-		else {
-			y = static_cast<SInt32>(0.5 + y *
-								static_cast<double>(dh - 1) / (sh - 1));
-		}
-		y += dy;
-		break;
-
-	case kTop:
-	case kBottom:
-		x -= sx;
-		if (x < 0) {
-			x = 0;
-		}
-		else if (x >= sw) {
-			x = dw - 1;
-		}
-		else {
-			x = static_cast<SInt32>(0.5 + x *
-								static_cast<double>(dw - 1) / (sw - 1));
-		}
-		x += dx;
-		break;
-
-	case kNoDirection:
-		assert(0 && "bad direction");
-		return NULL;
-	}
+	avoidJumpZone(dst, srcSide, x, y);
 
 	return dst;
 }
 
+void
+CServer::avoidJumpZone(IClient* dst, EDirection dir, SInt32& x, SInt32& y) const
+{
+	// we only need to avoid jump zones on the primary screen
+	if (dst != m_primaryClient) {
+		return;
+	}
+
+	const CString dstName(getName(dst));
+	SInt32 dx, dy, dw, dh;
+	dst->getShape(dx, dy, dw, dh);
+	float t = mapToFraction(dst, dir, x, y);
+	SInt32 z = getJumpZoneSize(dst);
+
+	// move in far enough to avoid the jump zone.  if entering a side
+	// that doesn't have a neighbor (i.e. an asymmetrical side) then we
+	// don't need to move inwards because that side can't provoke a jump.
+	switch (dir) {
+	case kLeft:
+		if (!m_config.getNeighbor(dstName, kRight, t, NULL).empty() &&
+			x > dx + dw - 1 - z)
+			x = dx + dw - 1 - z;
+		break;
+
+	case kRight:
+		if (!m_config.getNeighbor(dstName, kLeft, t, NULL).empty() &&
+			x < dx + z)
+			x = dx + z;
+		break;
+
+	case kTop:
+		if (!m_config.getNeighbor(dstName, kBottom, t, NULL).empty() &&
+			y > dy + dh - 1 - z)
+			y = dy + dh - 1 - z;
+		break;
+
+	case kBottom:
+		if (!m_config.getNeighbor(dstName, kTop, t, NULL).empty() &&
+			y < dy + z)
+			y = dy + z;
+		break;
+
+	case kNoDirection:
+		assert(0 && "bad direction");
+	}
+}
+
 bool
-CServer::isSwitchOkay(IClient* newScreen, EDirection dir, SInt32 x, SInt32 y)
+CServer::isSwitchOkay(IClient* newScreen, EDirection dir, SInt32 x, SInt32 y,
+								SInt32 xActive, SInt32 yActive)
 {
 	LOG((CLOG_DEBUG1 "try to leave \"%s\" on %s", getName(m_active).c_str(), CConfig::dirName(dir)));
 
@@ -709,6 +818,33 @@ CServer::isSwitchOkay(IClient* newScreen, EDirection dir, SInt32 x, SInt32 y)
 		preventSwitch = true;
 	}
 
+	// are we in a locked corner?  first check if screen has the option set
+	// and, if not, check the global options.
+	const CConfig::CScreenOptions* options =
+						m_config.getOptions(getName(m_active));
+	if (options == NULL || options->count(kOptionScreenSwitchCorners) == 0) {
+		options = m_config.getOptions("");
+	}
+	if (options != NULL && options->count(kOptionScreenSwitchCorners) > 0) {
+		// get corner mask and size
+		CConfig::CScreenOptions::const_iterator i =
+			options->find(kOptionScreenSwitchCorners);
+		UInt32 corners = static_cast<UInt32>(i->second);
+		i = options->find(kOptionScreenSwitchCornerSize);
+		SInt32 size = 0;
+		if (i != options->end()) {
+			size = i->second;
+		}
+
+		// see if we're in a locked corner
+		if ((getCorner(m_active, xActive, yActive, size) & corners) != 0) {
+			// yep, no switching
+			LOG((CLOG_DEBUG1 "locked in corner"));
+			preventSwitch = true;
+			stopSwitch();
+		}
+	}
+
 	// ignore if mouse is locked to screen and don't try to switch later
 	if (!preventSwitch && isLockedToScreen()) {
 		LOG((CLOG_DEBUG1 "locked to screen"));
@@ -840,6 +976,62 @@ CServer::isSwitchWaitStarted() const
 	return (m_switchWaitTimer != NULL);
 }
 
+UInt32
+CServer::getCorner(IClient* client, SInt32 x, SInt32 y, SInt32 size) const
+{
+	assert(client != NULL);
+
+	// get client screen shape
+	SInt32 ax, ay, aw, ah;
+	client->getShape(ax, ay, aw, ah);
+
+	// check for x,y on the left or right
+	SInt32 xSide;
+	if (x <= ax) {
+		xSide = -1;
+	}
+	else if (x >= ax + aw - 1) {
+		xSide = 1;
+	}
+	else {
+		xSide = 0;
+	}
+
+	// check for x,y on the top or bottom
+	SInt32 ySide;
+	if (y <= ay) {
+		ySide = -1;
+	}
+	else if (y >= ay + ah - 1) {
+		ySide = 1;
+	}
+	else {
+		ySide = 0;
+	}
+
+	// if against the left or right then check if y is within size
+	if (xSide != 0) {
+		if (y < ay + size) {
+			return (xSide < 0) ? kTopLeftMask : kTopRightMask;
+		}
+		else if (y >= ay + ah - size) {
+			return (xSide < 0) ? kBottomLeftMask : kBottomRightMask;
+		}
+	}
+
+	// if against the left or right then check if y is within size
+	if (ySide != 0) {
+		if (x < ax + size) {
+			return (ySide < 0) ? kTopLeftMask : kBottomLeftMask;
+		}
+		else if (x >= ax + aw - size) {
+			return (ySide < 0) ? kTopRightMask : kBottomRightMask;
+		}
+	}
+
+	return kNoCornerMask;
+}
+
 void
 CServer::stopRelativeMoves()
 {
@@ -889,6 +1081,7 @@ CServer::sendOptions(IClient* client) const
 	}
 
 	// send the options
+	client->resetOptions();
 	client->setOptions(optionsList);
 }
 
@@ -1020,7 +1213,7 @@ CServer::handleKeyDownEvent(const CEvent& event, void*)
 {
 	IPlatformScreen::CKeyInfo* info =
 		reinterpret_cast<IPlatformScreen::CKeyInfo*>(event.getData());
-	onKeyDown(info->m_key, info->m_mask, info->m_button);
+	onKeyDown(info->m_key, info->m_mask, info->m_button, info->m_screens);
 }
 
 void
@@ -1028,7 +1221,7 @@ CServer::handleKeyUpEvent(const CEvent& event, void*)
 {
 	IPlatformScreen::CKeyInfo* info =
 		 reinterpret_cast<IPlatformScreen::CKeyInfo*>(event.getData());
-	onKeyUp(info->m_key, info->m_mask, info->m_button);
+	onKeyUp(info->m_key, info->m_mask, info->m_button, info->m_screens);
 }
 
 void
@@ -1076,7 +1269,7 @@ CServer::handleWheelEvent(const CEvent& event, void*)
 {
 	IPlatformScreen::CWheelInfo* info =
 		reinterpret_cast<IPlatformScreen::CWheelInfo*>(event.getData());
-	onMouseWheel(info->m_wheel);
+	onMouseWheel(info->m_xDelta, info->m_yDelta);
 }
 
 void
@@ -1126,6 +1319,84 @@ CServer::handleClientCloseTimeout(const CEvent&, void* vclient)
 	delete client;
 }
 
+void
+CServer::handleSwitchToScreenEvent(const CEvent& event, void*)
+{
+	CSwitchToScreenInfo* info = 
+		reinterpret_cast<CSwitchToScreenInfo*>(event.getData());
+
+	CClientList::const_iterator index = m_clients.find(info->m_screen);
+	if (index == m_clients.end()) {
+		LOG((CLOG_DEBUG1 "screen \"%s\" not active", info->m_screen));
+	}
+	else {
+		jumpToScreen(index->second);
+	}
+}
+
+void
+CServer::handleSwitchInDirectionEvent(const CEvent& event, void*)
+{
+	CSwitchInDirectionInfo* info = 
+		reinterpret_cast<CSwitchInDirectionInfo*>(event.getData());
+
+	// jump to screen in chosen direction from center of this screen
+	SInt32 x = m_x, y = m_y;
+	IClient* newScreen = getNeighbor(m_active, info->m_direction, x, y);
+	if (newScreen == NULL) {
+		LOG((CLOG_DEBUG1 "no neighbor %s", CConfig::dirName(info->m_direction)));
+	}
+	else {
+		jumpToScreen(newScreen);
+	}
+}
+
+void
+CServer::handleLockCursorToScreenEvent(const CEvent& event, void*)
+{
+	CLockCursorToScreenInfo* info = (CLockCursorToScreenInfo*)event.getData();
+
+	// choose new state
+	bool newState;
+	switch (info->m_state) {
+	case CLockCursorToScreenInfo::kOff:
+		newState = false;
+		break;
+
+	default:
+	case CLockCursorToScreenInfo::kOn:
+		newState = true;
+		break;
+
+	case CLockCursorToScreenInfo::kToggle:
+		newState = !m_lockedToScreen;
+		break;
+	}
+
+	// enter new state
+	if (newState != m_lockedToScreen) {
+		m_lockedToScreen = newState;
+		LOG((CLOG_DEBUG "cursor %s current screen", m_lockedToScreen ? "locked to" : "unlocked from"));
+
+		m_primaryClient->reconfigure(getActivePrimarySides());
+		if (!isLockedToScreenServer()) {
+			stopRelativeMoves();
+		}
+	}
+}
+
+void
+CServer::handleFakeInputBeginEvent(const CEvent&, void*)
+{
+	m_primaryClient->fakeInputBegin();
+}
+
+void
+CServer::handleFakeInputEndEvent(const CEvent&, void*)
+{
+	m_primaryClient->fakeInputEnd();
+}
+
 void
 CServer::onClipboardChanged(IClient* sender, ClipboardID id, UInt32 seqNum)
 {
@@ -1133,7 +1404,7 @@ CServer::onClipboardChanged(IClient* sender, ClipboardID id, UInt32 seqNum)
 
 	// ignore update if sequence number is old
 	if (seqNum < clipboard.m_clipboardSeqNum) {
-		LOG((CLOG_INFO "ignored screen \"%s\" update of clipboard %d (missequenced)", clipboard.m_clipboardOwner.c_str(), id));
+		LOG((CLOG_INFO "ignored screen \"%s\" update of clipboard %d (missequenced)", getName(sender).c_str(), id));
 		return;
 	}
 
@@ -1221,33 +1492,45 @@ CServer::onScreensaver(bool activated)
 }
 
 void
-CServer::onKeyDown(KeyID id, KeyModifierMask mask, KeyButton button)
+CServer::onKeyDown(KeyID id, KeyModifierMask mask, KeyButton button,
+				const char* screens)
 {
 	LOG((CLOG_DEBUG1 "onKeyDown id=%d mask=0x%04x button=0x%04x", id, mask, button));
 	assert(m_active != NULL);
 
-	// handle command keys
-	if (onCommandKey(id, mask, true)) {
-		return;
-	}
-
 	// relay
-	m_active->keyDown(id, mask, button);
+	if (IKeyState::CKeyInfo::isDefault(screens)) {
+		m_active->keyDown(id, mask, button);
+	}
+	else {
+		for (CClientList::const_iterator index = m_clients.begin();
+								index != m_clients.end(); ++index) {
+			if (IKeyState::CKeyInfo::contains(screens, index->first)) {
+				index->second->keyDown(id, mask, button);
+			}
+		}
+	}
 }
 
 void
-CServer::onKeyUp(KeyID id, KeyModifierMask mask, KeyButton button)
+CServer::onKeyUp(KeyID id, KeyModifierMask mask, KeyButton button,
+				const char* screens)
 {
 	LOG((CLOG_DEBUG1 "onKeyUp id=%d mask=0x%04x button=0x%04x", id, mask, button));
 	assert(m_active != NULL);
 
-	// handle command keys
-	if (onCommandKey(id, mask, false)) {
-		return;
-	}
-
 	// relay
-	m_active->keyUp(id, mask, button);
+	if (IKeyState::CKeyInfo::isDefault(screens)) {
+		m_active->keyUp(id, mask, button);
+	}
+	else {
+		for (CClientList::const_iterator index = m_clients.begin();
+								index != m_clients.end(); ++index) {
+			if (IKeyState::CKeyInfo::contains(screens, index->first)) {
+				index->second->keyUp(id, mask, button);
+			}
+		}
+	}
 }
 
 void
@@ -1257,12 +1540,6 @@ CServer::onKeyRepeat(KeyID id, KeyModifierMask mask,
 	LOG((CLOG_DEBUG1 "onKeyRepeat id=%d mask=0x%04x count=%d button=0x%04x", id, mask, count, button));
 	assert(m_active != NULL);
 
-	// handle command keys
-	if (onCommandKey(id, mask, false)) {
-		onCommandKey(id, mask, true);
-		return;
-	}
-
 	// relay
 	m_active->keyRepeat(id, mask, count, button);
 }
@@ -1315,6 +1592,21 @@ CServer::onMouseMovePrimary(SInt32 x, SInt32 y)
 	m_active->getShape(ax, ay, aw, ah);
 	SInt32 zoneSize = getJumpZoneSize(m_active);
 
+	// clamp position to screen
+	SInt32 xc = x, yc = y;
+	if (xc < ax + zoneSize) {
+		xc = ax;
+	}
+	else if (xc >= ax + aw - zoneSize) {
+		xc = ax + aw - 1;
+	}
+	if (yc < ay + zoneSize) {
+		yc = ay;
+	}
+	else if (yc >= ay + ah - zoneSize) {
+		yc = ay + ah - 1;
+	}
+
 	// see if we should change screens
 	EDirection dir;
 	if (x < ax + zoneSize) {
@@ -1340,10 +1632,10 @@ CServer::onMouseMovePrimary(SInt32 x, SInt32 y)
 	}
 
 	// get jump destination
-	IClient* newScreen = getNeighbor(m_active, dir, x, y);
+	IClient* newScreen = mapToNeighbor(m_active, dir, x, y);
 
 	// should we switch or not?
-	if (isSwitchOkay(newScreen, dir, x, y)) {
+	if (isSwitchOkay(newScreen, dir, x, y, xc, yc)) {
 		// switch screen
 		switchScreen(newScreen, x, y, false);
 		return true;
@@ -1401,6 +1693,21 @@ CServer::onMouseMoveSecondary(SInt32 dx, SInt32 dy)
 	bool jump = true;
 	IClient* newScreen;
 	do {
+		// clamp position to screen
+		SInt32 xc = m_x, yc = m_y;
+		if (xc < ax) {
+			xc = ax;
+		}
+		else if (xc >= ax + aw) {
+			xc = ax + aw - 1;
+		}
+		if (yc < ay) {
+			yc = ay;
+		}
+		else if (yc >= ay + ah) {
+			yc = ay + ah - 1;
+		}
+
 		EDirection dir;
 		if (m_x < ax) {
 			dir = kLeft;
@@ -1457,10 +1764,10 @@ CServer::onMouseMoveSecondary(SInt32 dx, SInt32 dy)
 		}
 
 		// try to switch screen.  get the neighbor.
-		newScreen = getNeighbor(m_active, dir, m_x, m_y);
+		newScreen = mapToNeighbor(m_active, dir, m_x, m_y);
 
 		// see if we should switch
-		if (!isSwitchOkay(newScreen, dir, m_x, m_y)) {
+		if (!isSwitchOkay(newScreen, dir, m_x, m_y, xc, yc)) {
 			newScreen = m_active;
 			jump      = false;
 		}
@@ -1500,13 +1807,13 @@ CServer::onMouseMoveSecondary(SInt32 dx, SInt32 dy)
 }
 
 void
-CServer::onMouseWheel(SInt32 delta)
+CServer::onMouseWheel(SInt32 xDelta, SInt32 yDelta)
 {
-	LOG((CLOG_DEBUG1 "onMouseWheel %+d", delta));
+	LOG((CLOG_DEBUG1 "onMouseWheel %+d,%+d", xDelta, yDelta));
 	assert(m_active != NULL);
 
 	// relay
-	m_active->mouseWheel(delta);
+	m_active->mouseWheel(xDelta, yDelta);
 }
 
 bool
@@ -1705,3 +2012,61 @@ CServer::CClipboardInfo::CClipboardInfo() :
 {
 	// do nothing
 }
+
+
+//
+// CServer::CLockCursorToScreenInfo
+//
+
+CServer::CLockCursorToScreenInfo*
+CServer::CLockCursorToScreenInfo::alloc(State state)
+{
+	CLockCursorToScreenInfo* info =
+		(CLockCursorToScreenInfo*)malloc(sizeof(CLockCursorToScreenInfo));
+	info->m_state = state;
+	return info;
+}
+
+
+//
+// CServer::CSwitchToScreenInfo
+//
+
+CServer::CSwitchToScreenInfo*
+CServer::CSwitchToScreenInfo::alloc(const CString& screen)
+{
+	CSwitchToScreenInfo* info =
+		(CSwitchToScreenInfo*)malloc(sizeof(CSwitchToScreenInfo) +
+								screen.size());
+	strcpy(info->m_screen, screen.c_str());
+	return info;
+}
+
+
+//
+// CServer::CSwitchInDirectionInfo
+//
+
+CServer::CSwitchInDirectionInfo*
+CServer::CSwitchInDirectionInfo::alloc(EDirection direction)
+{
+	CSwitchInDirectionInfo* info =
+		(CSwitchInDirectionInfo*)malloc(sizeof(CSwitchInDirectionInfo));
+	info->m_direction = direction;
+	return info;
+}
+
+
+//
+// CServer::CScreenConnectedInfo
+//
+
+CServer::CScreenConnectedInfo*
+CServer::CScreenConnectedInfo::alloc(const CString& screen)
+{
+	CScreenConnectedInfo* info =
+		(CScreenConnectedInfo*)malloc(sizeof(CScreenConnectedInfo) +
+								screen.size());
+	strcpy(info->m_screen, screen.c_str());
+	return info;
+}
