@@ -29,6 +29,7 @@
 class CEventQueueTimer;
 class CPrimaryClient;
 class IClient;
+class CInputFilter;
 
 //! Synergy server
 /*!
@@ -36,6 +37,46 @@ This class implements the top-level server algorithms for synergy.
 */
 class CServer {
 public:
+	//! Lock cursor to screen data
+	class CLockCursorToScreenInfo {
+	public:
+		enum State { kOff, kOn, kToggle };
+
+		static CLockCursorToScreenInfo* alloc(State state = kToggle);
+
+	public:
+		State			m_state;
+	};
+
+	//! Switch to screen data
+	class CSwitchToScreenInfo {
+	public:
+		static CSwitchToScreenInfo* alloc(const CString& screen);
+
+	public:
+		// this is a C-string;  this type is a variable size structure
+		char			m_screen[1];
+	};
+
+	//! Switch in direction data
+	class CSwitchInDirectionInfo {
+	public:
+		static CSwitchInDirectionInfo* alloc(EDirection direction);
+
+	public:
+		EDirection		m_direction;
+	};
+
+	//! Screen connected data
+	class CScreenConnectedInfo {
+	public:
+		static CScreenConnectedInfo* alloc(const CString& screen);
+
+	public:
+		// this is a C-string;  this type is a variable size structure
+		char			m_screen[1];
+	};
+
 	/*!
 	Start the server with the configuration \p config and the primary
 	client (local screen) \p primaryClient.  The client retains
@@ -94,6 +135,14 @@ public:
 	*/
 	static CEvent::Type	getErrorEvent();
 
+	//! Get connected event type
+	/*!
+	Returns the connected event type.  This is sent when a client screen
+	has connected.  The event data is a \c CScreenConnectedInfo* that
+	indicates the connected screen.
+	*/
+	static CEvent::Type	getConnectedEvent();
+
 	//! Get disconnected event type
 	/*!
 	Returns the disconnected event type.  This is sent when all the
@@ -101,14 +150,31 @@ public:
 	*/
 	static CEvent::Type	getDisconnectedEvent();
 
-	//@}
-
-protected:
-	//! Handle special keys
+	//! Get switch to screen event type
 	/*!
-	Handles keys with special meaning.
+	Returns the switch to screen event type.  The server responds to this
+	by switching screens.  The event data is a \c CSwitchToScreenInfo*
+	that indicates the target screen.
 	*/
-	bool				onCommandKey(KeyID, KeyModifierMask, bool down);
+	static CEvent::Type	getSwitchToScreenEvent();
+
+	//! Get switch in direction event type
+	/*!
+	Returns the switch in direction event type.  The server responds to this
+	by switching screens.  The event data is a \c CSwitchInDirectionInfo*
+	that indicates the target direction.
+	*/
+	static CEvent::Type	getSwitchInDirectionEvent();
+
+	//! Get lock cursor event type
+	/*!
+	Returns the lock cursor event type.  The server responds to this
+	by locking the cursor to the active screen or unlocking it.  The
+	event data is a \c CLockCursorToScreenInfo*.
+	*/
+	static CEvent::Type	getLockCursorToScreenEvent();
+
+	//@}
 
 private:
 	// get canonical name of client
@@ -133,22 +199,46 @@ private:
 	void				switchScreen(IClient*,
 							SInt32 x, SInt32 y, bool forScreenSaver);
 
-	// lookup neighboring screen
-	IClient*			getNeighbor(IClient*, EDirection) const;
+	// jump to screen
+	void				jumpToScreen(IClient*);
+
+	// convert pixel position to fraction, using x or y depending on the
+	// direction.
+	float				mapToFraction(IClient*, EDirection,
+							SInt32 x, SInt32 y) const;
+
+	// convert fraction to pixel position, writing only x or y depending
+	// on the direction.
+	void				mapToPixel(IClient*, EDirection, float f,
+							SInt32& x, SInt32& y) const;
+
+	// returns true if the client has a neighbor anywhere along the edge
+	// indicated by the direction.
+	bool				hasAnyNeighbor(IClient*, EDirection) const;
+
+	// lookup neighboring screen, mapping the coordinate independent of
+	// the direction to the neighbor's coordinate space.
+	IClient*			getNeighbor(IClient*, EDirection,
+							SInt32& x, SInt32& y) const;
 
 	// lookup neighboring screen.  given a position relative to the
 	// source screen, find the screen we should move onto and where.
 	// if the position is sufficiently far from the source then we
 	// cross multiple screens.  if there is no suitable screen then
 	// return NULL and x,y are not modified.
-	IClient*			getNeighbor(IClient*, EDirection,
+	IClient*			mapToNeighbor(IClient*, EDirection,
+							SInt32& x, SInt32& y) const;
+
+	// adjusts x and y or neither to avoid ending up in a jump zone
+	// after entering the client in the given direction.
+	void				avoidJumpZone(IClient*, EDirection,
 							SInt32& x, SInt32& y) const;
 
 	// test if a switch is permitted.  this includes testing user
 	// options like switch delay and tracking any state required to
 	// implement them.  returns true iff a switch is permitted.
 	bool				isSwitchOkay(IClient* dst, EDirection,
-							SInt32 x, SInt32 y);
+							SInt32 x, SInt32 y, SInt32 xActive, SInt32 yActive);
 
 	// update switch state due to a mouse move at \p x, \p y that
 	// doesn't switch screens.
@@ -181,6 +271,11 @@ private:
 	// returns true iff the delay switch timer is started
 	bool				isSwitchWaitStarted() const;
 
+	// returns the corner (EScreenSwitchCornerMasks) where x,y is on the
+	// given client.  corners have the given size.
+	UInt32				getCorner(IClient*,
+							SInt32 x, SInt32 y, SInt32 size) const;
+
 	// stop relative mouse moves
 	void				stopRelativeMoves();
 
@@ -207,19 +302,26 @@ private:
 	void				handleSwitchWaitTimeout(const CEvent&, void*);
 	void				handleClientDisconnected(const CEvent&, void*);
 	void				handleClientCloseTimeout(const CEvent&, void*);
+	void				handleSwitchToScreenEvent(const CEvent&, void*);
+	void				handleSwitchInDirectionEvent(const CEvent&, void*);
+	void				handleLockCursorToScreenEvent(const CEvent&, void*);
+	void				handleFakeInputBeginEvent(const CEvent&, void*);
+	void				handleFakeInputEndEvent(const CEvent&, void*);
 
 	// event processing
 	void				onClipboardChanged(IClient* sender,
 							ClipboardID id, UInt32 seqNum);
 	void				onScreensaver(bool activated);
-	void				onKeyDown(KeyID, KeyModifierMask, KeyButton);
-	void				onKeyUp(KeyID, KeyModifierMask, KeyButton);
+	void				onKeyDown(KeyID, KeyModifierMask, KeyButton,
+							const char* screens);
+	void				onKeyUp(KeyID, KeyModifierMask, KeyButton,
+							const char* screens);
 	void				onKeyRepeat(KeyID, KeyModifierMask, SInt32, KeyButton);
 	void				onMouseDown(ButtonID);
 	void				onMouseUp(ButtonID);
 	bool				onMouseMovePrimary(SInt32 x, SInt32 y);
 	void				onMouseMoveSecondary(SInt32 dx, SInt32 dy);
-	void				onMouseWheel(SInt32 delta);
+	void				onMouseWheel(SInt32 xDelta, SInt32 yDelta);
 
 	// add client to list and attach event handlers for client
 	bool				addClient(IClient*);
@@ -289,6 +391,9 @@ private:
 	// current configuration
 	CConfig				m_config;
 
+	// input filter (from m_config);
+	CInputFilter*		m_inputFilter;
+
 	// clipboard cache
 	CClipboardInfo		m_clipboards[kClipboardEnd];
 
@@ -316,8 +421,15 @@ private:
 	// relative mouse move option
 	bool				m_relativeMoves;
 
+	// screen locking (former scroll lock)
+	bool				m_lockedToScreen;
+
 	static CEvent::Type	s_errorEvent;
+	static CEvent::Type	s_connectedEvent;
 	static CEvent::Type	s_disconnectedEvent;
+	static CEvent::Type	s_switchToScreen;
+	static CEvent::Type	s_switchInDirection;
+	static CEvent::Type s_lockCursorToScreen;
 };
 
 #endif
